"""
    cargo.field
    ~~~~~~~~~~~

    Implements a Field class which can be used
    to check individual fields within a given
    dataset and is used to validate them. Example
    usage::

        >>> from cargo.field import Field
        >>> f = Field(str).that.Is.Not.empty()
        >>> f.validate('')
        False

    They can be placed within Crate definitions
    and then you can visualize how it should be
    used in the ``cargo`` library::

        >>> from cargo.spec import Spec, Crate
        >>> class Model(Crate):
        ...     name = Field(str).that.Is.Not.empty()
        ...
        >>> Spec(Model).unpack({'name':'adam'})
        <__main__.Model object at 0x...>
        >>> Spec(Model).pack(_)
        {'name': 'adam'}
"""

from collections import defaultdict

_OP_MAP = {
    '==': lambda k,x: k == x,
    '>=': lambda k,x: k >= x,
    '<=': lambda k,x: k <= x,
    '!=': lambda k,x: k != x,
    '>':  lambda k,x: k > x,
    '<':  lambda k,x: k < x,
    '~=': lambda k,x: k == x or k == x-1 or k == x+1
    }
__all__ = ['Field']

class Field(object):
    """
    Represents a single field in the Crate
    specification, that can validate the
    equivalent within packed data. Usage::

        >>> from cargo.field import Field
        >>> f = Field(str)
        >>> f.validate('')
        True

    :param type_: A type to accept. Although
        it is possible to do a Field for a
        class, it is best to define another
        crate.
    """
    def __init__(self, type_):
        self.intended = type_
        self.flags = defaultdict(bool)
        self.chain = []

    @property
    def root(self):
        """
        Property to refer to the current
        field instance. Binded to the
        language chains in general.
        Usage example::

            >>> x = Field(str)
            >>> x.root == x
            True
        """
        return self

    @property
    def Not(self):
        """
        Toggle the internal not flag that will
        negate all validations succeeding this
        declaraion. Usage example::

            >>> Field(str).that.is.Not.empty()
            >>> Field(list).that.does.Not.contain(0)
        """
        value = True
        if self.flags['not']:
            value = False
        self.flags['not'] = value
        return self

    @property
    def has_default(self):
        """
        Tells whether a default value has been
        set in the current Field instance.
        Used internally by the Spec object::

            >>> Field(str).defaults_to('').has_default
            True
        """
        if 'default' in self.flags:
            return True
        return False

    def defaults_to(self, obj):
        """
        Declares that your Field will default
        to the given value if it is not present.
        Usage example::

            >>> Field(object).defaults_to([])

        :param obj: The object to default to.
        """
        self.flags['default'] = obj
        return self

    def satisfy(self, function):
        """
        Declares that the object that will be
        passed in for validation will satisfy
        (make the given function return a truthy
        result). Example::

            >>> Field(list).that.satisfy(lambda x: 0 in x)

        :param function: Any callable that can
            accept the object and return a true
            or false value.
        """
        if self.flags['not']:
            function = lambda x: not function(x)
        self.register(function)
        return self

    satisfies = satisfy

    def contains(self, value):
        """
        Validates that the passed in object will
        contain the given value when a membership
        test is being made, aka ``x in []``.
        Example usage::

            >>> Field(list).that.contains(0)

        You can also use the ``contain`` alias
        that refers to this function, if it seems
        more natural and readable.

        :param value: The value to check that
            is contained in the object.
        """
        function = lambda x: value in x
        if self.flags['not']:
            function = lambda x: value not in x
        self.register(function)
        return self

    contain = contains

    def value(self, operator, value):
        """
        Declares that objects validated by this
        field must return a truthy result when
        comparison operators are used on them,
        for example::

            >>> Field(int).that.has.value('>=', 2)

        The following operators are available:
         - '>=': greater or equal to
         - '<=': lesser or equal to
         - '==': equal to
         - '!=': not equal to
         - '~=': approximately equal to- a delta
           of 1 is tolerated.
         - '<': lesser than
         - '>': greater than

        :param operator: The operator string that
            can only be any of the above.
        :param value: The value.
        """
        operator = _OP_MAP[operator]
        function = lambda x: operator(x, value)
        if self.flags['not']:
            function = lambda x: not function(x)
        self.register(function)
        return self

    def implements(self, name):
        """
        Declares that the passed in object must
        implement a method under the given name,
        for example::

            >>> Field(object).that.implements('append')

        :param name: The name of the method.
        """
        self.attribute(name)
        function = lambda x: callable(getattr(x,name))
        if self.flags['not']:
            function = lambda x: not function(x)
        self.chain.append(function)
        return self

    def register(self, func):
        """
        Internal use only to append the function
        to the internal stack that will be used
        to validate objects. Usage example::

            >>> f = Field(object)
            >>> f.register(lambda x: True)
            >>> f.chain
            [<function <lambda> at 0x...>]

        :param func: The function to append to
            the end of the stack.
        """
        self.chain.append(func)

    def within(self, start, end):
        """
        Asserts that the passed in object must
        be within bounds of the two given
        values. Within meaning something like
        1 < x < 2. Usage example::

            >>> Field(float).that.Is.within(1,2)

        :param start: The minimum interval.
        :param end: The maximum interval.
        """
        function = lambda x: start < x < end
        if self.flags['not']:
            function = lambda x: not function(x)
        self.chain.append(function)
        return self

    def attribute(self, value):
        """
        Declares that the passed in object must
        have the given attribute. The attribute
        need not to be callable. Example::

            >>> Field(object).that.has.attribute('name')

        :param value: The name of the attribute.
        """
        function = lambda x: hasattr(x, value)
        if self.flags['not']:
            function = lambda x: not hasattr(x, value)
        self.chain.append(function)
        return self

    def _length(self, operator, length):
        operator = _OP_MAP[operator]
        function = lambda x: operator(len(x), length)
        if self.flags['not']:
            function = lambda x: not function(x)
        self.chain.append(function)
        return self

    def length(self, *args):
        """
        Declares comparison operators on the length
        of the passed in object. Accepts one or two
        arguments, example::

            >>> Field(list).With.length('~=', 1)
            >>> Field(list).With.length(1)

        Note that the two declarations are completely
        different because in the first one you can
        either accept 0, 1, or 2 length list, and in
        the second one you can only accept a 1 length
        list.

        :param operator: Optional, the operator to
            use as defined in ``Field.value``.
        :param value: The value to use.
        """
        if not args:
            raise TypeError('Expected 1 or more arguments')
        elif len(args) > 1:
            return self._length(*args)

        value = args[0]
        function = lambda x: len(x) == value
        if self.flags['not']:
            function = lambda x: len(x) != value
        self.chain.append(function)
        return self

    def empty(self):
        """
        Asserts that the passed in object's length
        must be 0 and zero only. Usually you will
        negate this declaration::

            >>> Field(list).that.is.Not.empty()
        """
        self.length(0)
        return self

    def validate(self, obj):
        """
        Validates a given object by first checking
        it's type and then running all registered
        functions upon the object, returning if
        they haven't passed. Example::

            >>> Field(list).validate({})
            False
            >>> Field(list).that.Is.empty().validate([])
            True

        :param obj: The object to validate.
        """
        if not isinstance(obj, self.intended):
            return False
        for function in self.chain:
            if not function(obj):
                return False
        return True

_NATURAL_KEYWORDS = ('Is','And','having','that','that','at','With','does')
for item in _NATURAL_KEYWORDS:
    setattr(Field, item, Field.root)

