"""
    cargo.spec
    ~~~~~~~~~~
    Functions and classes to help implement
    the cargo unpacking and packing interface.
    You can use this to enforce API requirements
    and turn responses into objects::

        >>> from cargo.spec import Spec, unpack
        >>> from cargo.field import Field
        >>> class Response(Spec):
        ...     error = Field(bool).defaults_to(None)
        ...     queue = Field(int).With.value('>=', 1)
        ...
        >>> g = API.call()
        >>> res = unpack(Response, g)
"""

from types import FunctionType
from cargo.field import Field


_INVALID = "key: '{0}' is invalid for model spec."
__all__ = ['Spec','pack','unpack']

class Spec(object):
    """
    Represents a schema specification
    that can be initialized by calling
    the constructor function with keyword
    arguments. Usage example::

        >>> class Model(Spec):
        ...     name = Field(str).With.length('>', 0)
        ...
        >>> model = Model(name='Eugene')
        >>> model.name
        'Eugene'
    """
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

def pack(spec, obj):
    """
    Packs an object into a dictionary
    based on the given specification.
    Does not attempt to validate any
    of the values though. Usage::

        >>> pack(Model, model)
        {'name':'Eugene'}

    :param spec: The Spec object/class.
    :param model: A model which is
        initialized by the same spec
        object.
    """
    final = {}
    for key, value in vars(spec).items():
        if not key.startswith('__'):
            cache = getattr(obj, key)
            if isinstance(value, Field):
                final[key] = cache

            elif isinstance(value, type) and issubclass(value, Spec):
                final[key] = pack(value, cache)
    return final

def unpack(spec, data):
    """
    Unpacks a dataset into a Python
    object. Usage example::

        >>> unpack(Model, {'name':'Eugene'})
        <__main__.Model at 0x...>

    :param spec: The specification
        class that contains definitions.
    :param data: The data that you want
        to convert into a Python class.
    """
    final = {}
    for key, value in vars(spec).items():
        if not key.startswith('__'):
            # cannot cache value beforehand because
            # value might be a function which is
            # unimplemented.
            if isinstance(value, Field):
                # check if we have a default value and
                # if we should go for the default value
                if key not in data and value.has_default:
                    thing = value.flags['default']
                else:
                    # maybe raise a KeyError
                    thing = data[key]
                if not value.validate(thing):
                    raise ValueError(_INVALID.format(key))
                final[key] = thing

            elif isinstance(value, type) and issubclass(value, Spec):
                thing = data[key]
                final[key] = unpack(value, thing)
    return spec(**final)

